/* adjustment.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "adjustment.hh"
#include <cassert>

using namespace gltk;

Adjustment::Adjustment(double val, double lo, double up, double st, double pg, double psz) :
	value(val),
	lower(lo),
	upper(up),
	step_increment(st),
	page_increment(pg),
	page_size(psz)
{}

Adjustment * Adjustment::create(double val, double lo, double up, double st, double pg, double ps)
{
	return new Adjustment(val, lo, up, st, pg, ps);
}

void Adjustment::configure(double val, double lo, double up, double st, double pg, double ps)
{
	struct _local {
		bool change;
		_local() : change(false) {}
		void md(double & var, double val)
		{
			if (var != val) {
				var = val;
				change |= true;
			}
		}
	} local;
	local.md(lower, lo);
	local.md(upper, up);
	local.md(step_increment, st);
	local.md(page_increment, pg);
	local.md(page_size, ps);
	if (local.change) changed.emit();
	if (value != val) {
		value = val;
		on_value_changed();
		value_changed.emit();
	}
}

double Adjustment::get_value() const
{
	return value;
}

void Adjustment::set_value(double v)
{
	if (v + page_size > upper) v = upper - page_size;
	if (v < lower) v = lower;
	if (v != value) {
		value = v;
		on_value_changed();
		value_changed.emit();
	}
}

void Adjustment::step_up()
{
	set_value(value + step_increment);
}

void Adjustment::step_down()
{
	set_value(value - step_increment);
}

void Adjustment::page_up()
{
	set_value(value + page_increment);
}

void Adjustment::page_down()
{
	set_value(value - page_increment);
}

void Adjustment::set_step_increment(double inc)
{
	step_increment = inc;
}

void Adjustment::set_page_increment(double inc)
{
	page_increment = inc;
}

void Adjustment::set_upper(double v)
{
	upper = v;
	assert(lower + page_size <= upper);
	if (value + page_size > upper) set_value(upper - page_size);
}

void Adjustment::set_lower(double v)
{
	lower = v;
	assert(lower + page_size <= upper);
	if (value < lower) set_value(lower);
}

void Adjustment::set_page_size(double ps)
{
	page_size = ps;
	assert(lower + page_size <= upper);
	if (value + page_size > upper) set_value(upper - page_size);
}

sigc::signal<void> & Adjustment::signal_changed()
{
	return changed;
}

sigc::signal<void> & Adjustment::signal_value_changed()
{
	return value_changed;
}
