/* box.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "box.hh"
#include "internal.hh"
#include <iostream>
using namespace gltk;

void Box::on_configure_event()
{
	int width = get_width();
	int height = get_height();
	if (hori) {
		int min = 0;
		int nat = 0;
		std::vector<Widget *>::const_iterator e = child_list.end();
		int cnt = 0;
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != e; i ++) {
			int m;
			int n;
			(* i)->get_preferred_width_for_height(width, m, n);
			min += m;
			nat += n;
			if (min > width) {
				// drop some widgets if height is not enough
				min -= m;
				nat -= n;
				e = i;
				break;
			}
			cnt ++;
		}
		int pad = width > nat && cnt > 1 ? (width - nat) / (cnt - 1) : 0;
		int ext = width - min - (cnt - 1) * pad; // extra space
		int x = 0;
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != e; i ++) {
			(* i)->get_preferred_width_for_height(height, min, nat);
			int w = min_of(nat, min + ext);
			(* i)->size_allocate(x, 0, w, height);
			x += w + pad;
			ext -= (w - min);
		}
		while (e != child_list.end()) { // crushed widgets (zero width)
			(* e)->size_allocate(width, 0, 0, height);
			e ++;
		}
	}
	else {
		int min = 0;
		int nat = 0;
		std::vector<Widget *>::const_iterator e = child_list.end();
		int cnt = 0;
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != e; i ++) {
			int m;
			int n;
			(* i)->get_preferred_height_for_width(width, m, n);
			min += m;
			nat += n;
			if (min > height) {
				// drop some widgets if height is not enough
				min -= m;
				nat -= n;
				e = i;
				break;
			}
			cnt ++;
		}
		int pad = height > nat && cnt > 1 ? (height - nat) / (cnt - 1) : 0;
		int ext = height - min - (cnt - 1) * pad; // extra space
		int y = 0;
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != e; i ++) {
			(* i)->get_preferred_height_for_width(width, min, nat);
			int h = min_of(nat, min + ext);
			(* i)->size_allocate(0, y, width, h);
			y += h + pad;
			ext -= (h - min);
		}
		while (e != child_list.end()) { // crushed widgets (zero height)
			(* e)->size_allocate(0, height, width, 0);
			e ++;
		}
	}
}

void Box::get_preferred_width(int & min, int & nat) const
{
	min = 0;
	nat = 0;
	if (hori) {
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != child_list.end(); i ++) {
			int m;
			int n;
			(* i)->get_preferred_width(m, n);
			min += m;
			nat += n;
		}
	}
	else {
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != child_list.end(); i ++) {
			int m;
			int n;
			(* i)->get_preferred_width(m, n);
			min = max_of(min, m);
			nat = max_of(nat, n);
		}
	}
	nat += border_width;
}

void Box::get_preferred_height(int & min, int & nat) const
{
	min = 0;
	nat = 0;
	if (hori) {
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != child_list.end(); i ++) {
			int m;
			int n;
			(* i)->get_preferred_height(m, n);
			min = max_of(min, m);
			nat = max_of(nat, n);
		}
	}
	else {
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != child_list.end(); i ++) {
			int m;
			int n;
			(* i)->get_preferred_height(m, n);
			min += m;
			nat += n;
		}
	}
	nat += border_width;
}

void Box::get_preferred_width_for_height(int height, int & min, int & nat) const
{
	min = 0;
	nat = 0;
	if (hori) {
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != child_list.end(); i ++) {
			int m;
			int n;
			(* i)->get_preferred_width_for_height(height, m, n);
			min += m;
			nat += n;
		}
	}
	else {
		// How do we distribute the height to all children? This could be complicated.
		std::vector<Widget *>::const_iterator e = child_list.end();
		int cnt = 0;
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != e; i ++) {
			int m;
			int n;
			(* i)->get_preferred_height(m, n);
			min += m;
			nat += n;
			if (min > height) {
				// drop some widgets if height is not enough
				min -= m;
				nat -= n;
				e = i;
				break;
			}
			cnt ++;
		}
		int ext = height - min; // extra space
		min = 0;
		nat = 0;
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != e; i ++) {
			int mh;
			int nh;
			(* i)->get_preferred_height(mh, nh);
			int h = min_of(nh, mh + ext);
			int mw;
			int nw;
			(* i)->get_preferred_width_for_height(h, mw, nw);
			ext -= (h - mh);
			min = max_of(min, mw);
			nat = max_of(nat, nw);
		}
	}
	nat += border_width;
}

void Box::get_preferred_height_for_width(int width, int & min, int & nat) const
{
	if (hori) {
		// How do we distribute the width to all children? This could be complicated.
		min = 0;
		nat = 0;
		std::vector<Widget *>::const_iterator e = child_list.end();
		int cnt = 0;
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != e; i ++) {
			int m;
			int n;
			(* i)->get_preferred_width(m, n);
			min += m;
			nat += n;
			if (min > width) {
				// drop some widgets if width is not enough
				min -= m;
				nat -= n;
				e = i;
				break;
			}
			cnt ++;
		}
		int ext = width - min; // extra space
		min = 0;
		nat = 0;
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != e; i ++) {
			int mw;
			int nw;
			(* i)->get_preferred_width(mw, nw);
			int w = min_of(nw, mw + ext);
			int mh;
			int nh;
			(* i)->get_preferred_height_for_width(w, mh, nh);
			ext -= (w - mw);
			min = max_of(min, mh);
			nat = max_of(nat, nh);
		}
	}
	else {
		min = 0;
		nat = 0;
		for (std::vector<Widget *>::const_iterator i = child_list.begin(); i != child_list.end(); i ++) {
			int m;
			int n;
			(* i)->get_preferred_height_for_width(width, m, n);
			min += m;
			nat += n;
		}
	}
}

Box::Box(Orientation ori, int sp) :
	Orientable(ori),
	border_width(0),
	spacing(sp)
{}

void Box::pack_start(Widget * child)
{
	add(child);
}

void Box::set_border_width(unsigned bw)
{
	border_width = bw;
}
