/* entry.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "entry.hh"
#include "glut_inc.hh"
#include "internal.hh"
#include <algorithm>
#include <iostream>
#include <cctype>
#include <cassert>
using namespace gltk;

void Entry::on_draw()
{
	int w = get_width();
	int h = get_height();
	glColor3f(1.0, 1.0, 1.0);
	glBegin(GL_QUADS);
	glVertex2i(0, 0);
	glVertex2i(w, 0);
	glVertex2i(w, h);
	glVertex2i(0, h);
	glEnd();
	glut_draw_sunken_box(0, 0, w, h);

	// draw highlighted marked region
	int cx = - 1;
	int ex = (h - glut_get_font_height(font)) / 2; // top & bottom margin
	if (marked) {
		if (cx < 0) cx = find_x(cursor);
		int mx = find_x(mark);
		glColor3f(0.8, 0.8, 1.0);
		glBegin(GL_QUADS);
		glVertex2i(mx, ex);
		glVertex2i(cx, ex);
		glVertex2i(cx, ex + glut_get_font_height(font));
		glVertex2i(mx, ex + glut_get_font_height(font));
		glEnd();
	}
	if (vstart) { // need to draw starting arrow
		glColor3f(0.5, 0.25, 0.25);
		glBegin(GL_TRIANGLES);
		glVertex2f(bw + as - 1, ex);
		glVertex2f(bw + as - 1, ex + glut_get_font_height(font));
		glVertex2f(bw - 1, h / 2.0);
		glEnd();
	}
	if (visibility) glColor3f(0, 0, 0);
	else glColor3f(1.0, 0.4, 0.4);
	int xs = vstart ? bw + as : bw;
	int mx = w - bw;
	glRasterPos2i(xs, ex + glut_get_font_ascent(font));
	bool check = false;
	end_arrow = false;
	for (size_t i = vstart; i < text.size(); i ++) {
		char c = visibility ? text[i] : '*';
		xs += glutBitmapWidth(font, c);
		if (xs > mx - as && ! check) {
			// can we fit the whole thing without the arrow?
			size_t j = i;
			int xss = xs;
			check = true;
			do {
				if (xss > mx) { // No, can not fit!
					// draw ending arrow
					glColor3f(0.5, 0.25, 0.25);
					glBegin(GL_TRIANGLES);
					glVertex2f(w - bw - as + 1, ex);
					glVertex2f(w - bw - as + 1, ex + glut_get_font_height(font));
					glVertex2f(w - bw + 1, h / 2.0);
					glEnd();
					end_arrow = true;
					break;
				}
				if (++ j > text.size()) break;
				xss += glutBitmapWidth(font, visibility ? text[j] : '*');
			} while (true);
			if (end_arrow) break;
		}
		glutBitmapCharacter(font, c);
	}

	// drawing cursor
	if (focused) {
		if (cx < 0) cx = find_x(cursor);
		glBegin(GL_LINES);
		if (editable) glColor3f(0, 0, 1.0);
		else glColor3f(0.6, 0.6, 0.0);
		glVertex2f(0.5 + cx, ex);
		glVertex2f(0.5 + cx, ex + glut_get_font_height(font));
		glEnd();
	}
}

void Entry::on_button_event(int button, int state, int x, int y)
{
	if (! focused) {
		if (get_focus()) {
			focused = true;
			queue_draw();
		}
	}
	if (button != GLUT_LEFT_BUTTON) return;
	if (state == GLUT_DOWN) {
		set_grabbing();
		cursor = find_position(x);
		marked = true;
		mark = cursor;
	}
	else {
		set_grabbing(false);
	}
	queue_draw();
}

void Entry::on_motion_event(int x, int y)
{
	if (! marked) return;
	if (x < 0 || x > get_width() - bw - 2) return;
	size_t p = find_position(x);
	if (p != cursor) {
		cursor = p;
		queue_draw();
	}
}

bool Entry::on_key_event(unsigned char key)
{
	if (isprint(key)) {
		if (editable) {
			if (marked) delete_marked();
			text.insert(cursor, 1, char(key));
			cursor ++;
			adjust_vstart();
		}
		marked = false;
	}
	else switch (key) {
	case 0x8: // Backspace
		if (editable) {
			if (marked) delete_marked();
			else if (cursor > 0) {
				cursor --;
				if (vstart > cursor) vstart = cursor;
				text.erase(cursor, 1);
			}
		}
		break;
	case 0xb: // ^K
		if (editable) {
			if (marked) delete_marked();
			else if (cursor < text.size()) text.erase(cursor);
		}
		break;
	case 0xd: // Enter
		activate.emit();
		select_all();
		return true; // do not unmark
	case 0x7f: // Delete
		if (editable) {
			if (marked) delete_marked();
			else if (cursor < text.size()) text.erase(cursor, 1);
		}
		break;
	default:
		return false;
	}
	marked = false;
	queue_draw();
	return true;
}

bool Entry::on_special_event(int key)
{
	if (! marked && glutGetModifiers() == GLUT_ACTIVE_SHIFT) {
		mark = cursor;
		marked = true;
	}
	switch (key) {
	case GLUT_KEY_LEFT:
		if (cursor > 0) {
			cursor --;
			if (vstart > cursor) vstart = cursor;
		}
		break;
	case GLUT_KEY_RIGHT:
		if (cursor < text.size()) {
			cursor ++;
			adjust_vstart();
		}
		break;
	case GLUT_KEY_HOME:
		vstart = 0;
		cursor = 0;
		break;
	case GLUT_KEY_END:
		cursor = text.size();
		adjust_vstart();
		break;
	default:
		return false;
	}
	if (glutGetModifiers() != GLUT_ACTIVE_SHIFT) marked = false;
	queue_draw();
	return true;
}

bool Entry::focus_make()
{
	focused = true;
	select_all();
	return true;
}

void Entry::focus_end()
{
	if (focused) {
		focused = false;
		stop.emit();
		select_none();
		queue_draw();
	}
}

void Entry::adjust_vstart()
{
	int aw = get_width() - bw * 2; // available width
	if (visibility)	{
		int w = 0;
		// find text width from vstart to cursor
		for (size_t i = vstart; i < cursor; i ++) w += glutBitmapWidth(font, text[i]);
		int ce = 0; // cursor to end size
		for (size_t j = cursor; j < text.size() && ce <= as; j ++) {
			ce += glutBitmapWidth(font, text[j]);
		}
		if (vstart) w += as;
		while (w > aw - as && w + ce > aw) {
			w -= glutBitmapWidth(font, text[vstart]);
			if (! vstart) w += as;
			vstart ++;
		}
	}
	else {
		int w = glutBitmapWidth(font, '*') * (cursor - vstart);
		if ((vstart ? w + as : w) > aw) {
			vstart = cursor - (aw - as) / glutBitmapWidth(font, '*');
			if (vstart < 0) vstart = 0;
		}
	}
}

int Entry::find_x(size_t p)
{
	int x = vstart ? bw + as : bw;
	if (p <= vstart) return x;
	if (p > text.size()) p = text.size();
	if (visibility) for (size_t i = vstart; i < p; i ++) x += glutBitmapWidth(font, text[i]);
	else x += glutBitmapWidth(font, '*') * (p - vstart);
	return min_of(x, get_width() - bw);
}


size_t Entry::find_position(int x)
{
	int w = vstart ? bw + as : bw;
	int mx = get_width() - (end_arrow ? bw + as : bw);
	for (size_t i = vstart; i < text.size(); i ++) {
		int cw = glutBitmapWidth(font, text[i]);
		if (w + cw / 2 > x) return i;
		w += cw;
		if (w > mx) return i;
	}
	return text.size();
}

void Entry::delete_marked()
{
	if (mark < cursor) {
		text.erase(mark, cursor - mark);
		cursor = mark;
		if (vstart > cursor) vstart = cursor;
	}
	else text.erase(cursor, mark - cursor);
	marked = false;
}

Entry::Entry() :
	focused(false),
	editable(true),
	visibility(true),
	cursor(0),
	vstart(0),
	marked(false),
	font(GLUT_BITMAP_HELVETICA_12),
	width_chars(0)
{
	bw = glut_sunken_box_width() + 1;
	as = 5;
	end_arrow = false;
}

void Entry::get_preferred_height(int & min, int & nat) const
{
	int h = glut_get_font_height(font);
	min = h + bw * 2;
	nat = h + bw * 2 + 4;
}

void Entry::get_preferred_width(int & min, int & nat) const
{
	if (width_chars) {
		nat = width_chars * glutBitmapWidth(font, 'W') + bw * 2;
		min = nat;
	}
	else {
		min = 10 + bw * 2;
		nat = 100 + bw * 2;
	}
}

void Entry::set_text(std::string const & t)
{
	text = t;
	if (cursor > t.size()) {
		cursor = t.size();
		if (vstart > cursor) vstart = cursor;
	}
	if (marked && mark > t.size()) mark = t.size();
	queue_draw();
}

std::string const & Entry::get_text() const
{
	return text;
}

size_t Entry::get_text_length() const
{
	return text.size();
}

void Entry::select_all()
{
	mark = text.size();
	marked = true;
	cursor = 0;
	vstart = 0;
	queue_draw();
}

void Entry::select_none()
{
	marked = false;
	cursor = 0;
	vstart = 0;
	queue_draw();
}

void Entry::set_max_length(size_t length)
{
	max_length = length;
}

void Entry::set_width_chars(int n_chars)
{
	width_chars = n_chars;
}

void Entry::set_editable(bool v)
{
	editable = v;
}

void Entry::set_visibility(bool v)
{
	if (v != visibility) {
		visibility = v;
		if (get_realized()) adjust_vstart();
	}
}

sigc::signal<void> & Entry::signal_activate()
{
	return activate;
}

sigc::signal<void> & Entry::signal_stop()
{
	return stop;
}
