/* filebrowser.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "filebrowser.hh"
#include "scrollbar.hh"
#include "glut_inc.hh"
#include "internal.hh"
#include <iostream>
#include <algorithm>
#include <vector>
#include <list>

#ifdef __GNUC__
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#elif defined _WIN32
#include <windows.h>
#endif

using namespace gltk;

namespace {
	struct FileItem :
		ViewableList::ItemBase
	{
		enum Type {
			FILE,
			DIRECTORY
		};

		std::string name;
		Type type;
		FileItem(std::string const & n, Type t) : name(n), type(t) {}

		// How to compare file items
		bool operator<(FileItem const & f)
		{
			if (type != f.type) return type == DIRECTORY; // directory first
			return name < f.name;
		}
	};

	struct RenderFileItem :
		public ViewableList::RenderBase
	{
		void * font;
		int get_height()
		{
			return glut_get_font_height(font) + 2;
		}

		void render(ViewableList::ItemBase * item, int x, int y, int width, int height)
		{
			if (FileItem * i = dynamic_cast<FileItem *>(item)) {
				if (i->type == FileItem::DIRECTORY) glColor3f(0, 0.8, 0);
				else glColor3f(0, 0, 0);
				glRasterPos2i(x + 1, y + 1 + glut_get_font_ascent(font));
				glut_draw_string(font, i->name, width - 2);
			}
			else throw Error("unexpected ItemBase is not a FileItem");
		}

		RenderFileItem() :
			font(GLUT_BITMAP_HELVETICA_12)
		{}
	};

	class FileList :
		public ViewableList
	{
		// storage implementation
		std::list<FileItem> filelist;
		std::vector<std::list<FileItem>::iterator> fl_index;

		// overrides
		size_t size() const
		{
			return fl_index.size();
		}

		ItemBase * get_item(size_t i)
		{
			return & (* fl_index[i]);
		}

		void read_filelist()
		{
			filelist.clear();
#ifdef __GNUC__
			DIR * dir;
			struct dirent * dirp;
			struct stat dr;
			if (! (dir = opendir("."))) throw Error("Fail to opendir .");
			while ((dirp = readdir(dir))) {   // open directory
				filelist.push_back(FileItem(dirp->d_name, ! stat(dirp->d_name, & dr) && S_ISDIR(dr.st_mode) ? FileItem::DIRECTORY : FileItem::FILE));
			}
			closedir(dir);
#elif defined _WIN32
			WIN32_FIND_DATA fdata;
			HANDLE fhndl = FindFirstFile("*", & fdata);
			if (fhndl == INVALID_HANDLE_VALUE) throw Error("Fail to list current directory");
			do {
				filelist.push_back(FileItem(fdata.cFileName, fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ? FileItem::DIRECTORY : FileItem::FILE));
			} while (FindNextFile(fhndl, & fdata) != 0);
			if (GetLastError() == ERROR_NO_MORE_FILES) FindClose(fhndl);
			else throw Error("Error reading current directory ");
#endif
			filelist.sort();
		}

		void index_filelist()
		{
			fl_index.clear();
			for (std::list<FileItem>::iterator i = filelist.begin(); i != filelist.end(); i ++) {
				fl_index.push_back(i);
			}
		}

	public:
		FileList()
		{
			read_filelist();
			index_filelist();
		}

		void item_selected(int item)
		{
			FileItem & fi = * fl_index[item];
			if (fi.type == FileItem::DIRECTORY) {
#ifdef __GNUC__
				chdir(fi.name.c_str());
#elif defined _WIN32
				SetCurrentDirectory(fi.name.c_str());
#endif
				read_filelist();
				index_filelist();
				changed.emit();
			}
			else select_file.emit(fi.name);
		}
		sigc::signal<void, std::string> select_file;
	};
}

FileBrowser::FileBrowser() :
	Box(ORIENTATION_HORIZONTAL)
{
	FileList * fl = new FileList;
	ListView * lv = new ListView;
	Scrollbar * sb = new Scrollbar(lv, ORIENTATION_VERTICAL);
	RenderFileItem * rfi = new RenderFileItem;
	pack_start(lv);
	pack_start(sb);
	lv->set_list(fl);
	lv->add_render(rfi);
	lv->signal_selected().connect(sigc::mem_fun(fl, & FileList::item_selected));
	list = fl;
}

void FileBrowser::set_path(std::string const & path)
{
#ifdef __GNUC__
	chdir(path.c_str());
#elif defined _WIN32
	SetCurrentDirectory(path.c_str());
#endif
}

std::string FileBrowser::get_path() const
{
	std::string path;
#ifdef __GNUC__
	size_t len = 32;
	char * buf = new char [len];
	while (! getcwd(buf, len)) {
		delete [] buf;
		len *= 2;
		buf = new char[len];
	}
	path = buf;
#elif defined _WIN32
	size_t len = GetCurrentDirectory(0, 0);
	char * buf = new char [len + 1];
	GetCurrentDirectory(len + 1, buf);
	path = buf;
	delete [] buf;
#endif
	return path;
}

sigc::signal<void, std::string> & FileBrowser::signal_selected()
{
	FileList * fl = dynamic_cast<FileList *>(list);
	return fl->select_file;
}
