/* spinbutton.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "spinbutton.hh"
#include "shape.hh"
#include "timeout.hh"
#include "glut_inc.hh"
#include "internal.hh"
#include <sstream>
#include <iomanip>
#include <cmath>
#include <cassert>

using namespace gltk;

void SpinButton::on_configure_event()
{
	int height = get_height();
	int width = get_width();
	int h2 = height / 2;
	int h = min_of(h2, 7);
	if (width < 10) {
		tx.size_allocate(0, 0, width, height);
		up.size_allocate(width, h2 - h, 0, h);
		dn.size_allocate(width, h2, 0, h);
	}
	else {
		int x = width - 10;
		tx.size_allocate(0, 0, x, height);
		up.size_allocate(x, h2 - h, 10, h);
		dn.size_allocate(x, h2, 10, h);
	}
}

void SpinButton::on_draw()
{
	draw_children();
}

void SpinButton::on_value_changed()
{
	std::ostringstream ss;
	ss << std::fixed << std::setprecision(digits) << get_value();
	tx.set_text(ss.str());
	queue_draw();
}

void SpinButton::on_button_event(int button, int state, int x, int y)
{
	if (button == GLUT_LEFT_BUTTON) {
		if (grabber) {
			assert(state == GLUT_UP);
			grabber->button_event(button, state, x, y);
			assert(! grabber);
			if (held) {
				held_timeout.disconnect();
				held = false;
			}
		}
		else if (Widget * w = get_child(x, y)) {
			w->button_event(button, state, x, y);
		}
		if (grabber && ! held && grabber != & tx) { // a Button became held
			spin_interval = spin_wait / 2;
			climb_value = step_increment;
			held_timeout = SignalTimeout::connect(sigc::mem_fun(this, & SpinButton::button_held), spin_wait);
			held = true;
			spinning = false;
		}
	}
	else if (button == 3) { // FreeGLUT, wheel up
		if (state == GLUT_DOWN) pushed_up();
	}
	else if (button == 4) { // FreeGLUT, wheel down
		if (state == GLUT_DOWN) pushed_dn();
	}
}

bool SpinButton::on_special_event(int key)
{
	if (! focus) return false;
	if (focus == & tx) {
		switch (key) {
		case GLUT_KEY_UP:
			pushed_up();
			break;
		case GLUT_KEY_DOWN:
			pushed_dn();
			break;
		case GLUT_KEY_PAGE_UP:
			page_up();
			break;
		case GLUT_KEY_PAGE_DOWN:
			page_down();
			break;
		default:
			return focus->special_event(key);
		}
		tx.select_all();
		return true;
	}
	return focus->special_event(key);
}

void SpinButton::value_enter()
{
	std::istringstream s(tx.get_text());
	double v;
	s >> v;
	if (snap_to_ticks) v = step_increment * round_int(v / step_increment);
	if (v != get_value()) set_value(v);
	else {
		std::ostringstream ss;
		ss << std::fixed << std::setprecision(digits) << get_value();
		tx.set_text(ss.str());
		queue_draw();
	}
}

void SpinButton::value_stop()
{
	std::ostringstream ss;
	ss << std::fixed << std::setprecision(digits) << get_value();
	tx.set_text(ss.str());
}

bool SpinButton::button_held()
{
	if (! grabber) { // button released
		held = false;
		return false;
	}
	spinning = true;
	double v = get_value();
	if (grabber == & up) {
		v += climb_value;
		if (wrap && v > upper) v = lower;
	}
	else if (grabber == & dn) {
		v -= climb_value;
		if (wrap && v < lower) v = upper;
	}
	else throw Error("Unknown grabber...");

	if (snap_to_ticks) v = step_increment * round_int(v / step_increment);
	set_value(v);
	spin_interval -= (spin_interval / 3);
	spin_interval = max_of(spin_interval, spin_wait / 10);
	climb_value += climb_value * climb_rate / 10;
	if (climb_value > page_increment) climb_value = (page_increment + climb_value) / 2;
	held_timeout = SignalTimeout::connect(sigc::mem_fun(this, & SpinButton::button_held), spin_interval);
	return false;
}

void SpinButton::pushed_up()
{
	double v = get_value() + step_increment;
	if (wrap && v > upper) v = lower;
	if (snap_to_ticks) v = step_increment * round_int(v / step_increment);
	set_value(v);
}

void SpinButton::pushed_dn()
{
	double v = get_value() - step_increment;
	if (wrap && v < lower) v = upper;
	if (snap_to_ticks) v = step_increment * round_int(v / step_increment);
	set_value(v);
}

SpinButton::SpinButton(double c_rate, unsigned dgts) :
	Adjustment(0, 0, 10),
	up(new Shape(Shape::UP_TRIANGLE, 8, 4)),
	dn(new Shape(Shape::DOWN_TRIANGLE, 8, 4)),
	spin_wait(700),
	held(false),
	spinning(false),
	climb_rate(c_rate),
	digits(dgts),
	snap_to_ticks(false),
	wrap(false)
{
	add(& tx);
	add(& up);
	add(& dn);
	// reduced box_width for more compact arrows
	up.set_reserved(3, 3);
	dn.set_reserved(3, 3);
	up.set_box_width(1);
	dn.set_box_width(1);
	up.signal_pushed().connect(sigc::mem_fun(this, & SpinButton::pushed_up));
	dn.signal_pushed().connect(sigc::mem_fun(this, & SpinButton::pushed_dn));
	std::ostringstream s;
	s << std::fixed << std::setprecision(digits) << get_value();
	tx.set_text(s.str());
	tx.signal_activate().connect(sigc::mem_fun(this, & SpinButton::value_enter));
	tx.signal_stop().connect(sigc::mem_fun(this, & SpinButton::value_stop));
}

void SpinButton::get_preferred_width(int & min, int & nat) const
{
	tx.get_preferred_width(min, nat);
	min += 10;
	nat += 10;
}

void SpinButton::get_preferred_height(int & min, int & nat) const
{
	tx.get_preferred_height(min, nat);
	min = max_of(min, 10);
	nat = max_of(nat, 16);
}

int SpinButton::get_value_as_int() const
{
	int v = get_value();
	return v;
}

unsigned SpinButton::get_digits() const
{
	return digits;
}

void SpinButton::set_digits(unsigned d)
{
	digits = d;
	std::ostringstream ss;
	ss << std::fixed << std::setprecision(digits) << get_value();
	tx.set_text(ss.str());
	tx.select_none();
	queue_draw();
}

void SpinButton::set_snap_to_ticks(bool snap)
{
	snap_to_ticks = snap;
	if (snap) {
		double v = step_increment * round_int(get_value() / step_increment);
		set_value(v);
	}
}

void SpinButton::set_increments(double step, double page)
{
	set_step_increment(step);
	set_page_increment(page);
	if (snap_to_ticks) {
		double v = step_increment * round_int(get_value() / step_increment);
		set_value(v);
	}
}

void SpinButton::set_range(double min, double max)
{
	set_upper(max);
	set_lower(min);
}

void SpinButton::set_wrap(bool w)
{
	wrap = w;
}

void SpinButton::set_width_chars(int n_chars)
{
	tx.set_width_chars(n_chars);
}
