/* window.cc
 * 
 * Copyright (C) 2011 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This file is part of gltk.
 * 
 * gltk is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with gltk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include "window.hh"
#include "glut_inc.hh"
#include "internal.hh"
#include <iostream>
#include <cassert>

using namespace gltk;

size_t Window::visible_count = 0;

void Window::on_display_func()
{
	if (to_reconf) {
		size_allocate(0, 0, win_width, win_height);
		to_reconf = false;
	}
	// setup 2D drawing
	glDisable(GL_LIGHTING);
	glViewport(0, 0, win_width, win_height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, win_width, win_height, 0, 1, - 1); // (0, 0) @ upper-left
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glColor3f(0.9, 0.9, 0.9);
	glBegin(GL_QUADS);
	glVertex2d(0, 0);
	glVertex2d(0, win_height);
	glVertex2d(win_width, win_height);
	glVertex2d(win_width, 0);
	glEnd();
	draw_children();
	glutSwapBuffers();
}

void Window::on_reshape_func(int width, int height)
{
	win_width = width;
	win_height = height;
	size_allocate(0, 0, width, height);
	to_reconf = false;
}

void Window::on_mouse_func(int button, int state, int x, int y)
{
	button_event(button, state, x, y);
}

void Window::on_motion_func(int x, int y)
{
	motion_event(x, y);
}

void Window::on_passive_func(int x, int y)
{
	passive_event(x, y);
}

void Window::on_keyboard_func(unsigned char key, int x, int y)
{
	if (key == 0x9) { // tab
		((Widget *) this)->focus_next();
		return;
	}
	if (! keyboard_event(key)) std::cerr << "unhandled key: 0x" << std::hex << int(key) << '\n';
}

void Window::on_special_func(int key, int x, int y)
{
	if (! special_event(key)) std::cerr << "unhandled special: 0x" << std::hex << key << '\n';
}

void Window::on_create()
{
	assert(get_win_id());
	realize_children();
}

void Window::on_destroy()
{
	unrealize_children();
	if (visible) {
		visible = false;
		visible_count --;
	}
}

void Window::on_add(Widget * child)
{
	Bin::on_add(child);
	child->set_window(this);
}

bool Window::get_focus()
{
	return true;
}

Window::Window() :
	visible(false),
	to_reconf(false),
	default_w(0),
	default_h(0)
{
	set_events(
		DISPLAY_MASK |
		RESHAPE_MASK |
		MOUSE_MASK |
		MOTION_MASK |
		PASSIVE_MASK |
		KEYBOARD_MASK |
		SPECIAL_MASK
	);
}

Window::~Window()
{
	if (visible) visible_count --;
}

void Window::set_default_size(int w, int h)
{
	default_w = w;
	default_h = h;
}

void Window::set_title(std::string const & txt)
{
	title = txt;
	if (get_realized()) {
		glutSetWindow(get_win_id());
		glutSetWindowTitle(title.c_str());
	}
}

void Window::resize(int w, int h)
{
	if (get_win_id()) glutWindow::resize(w, h);
}

void Window::refresh()
{
	redisplay();
}

void Window::reconfigure()
{
	to_reconf = true;
	redisplay();
}

void Window::make_current()
{
	set_current();
}

void Window::show()
{
	if (int wid = get_win_id()) glutSetWindow(wid);
	else {
		int mw;
		int nw;
		get_preferred_width(mw, nw);
		if (default_w) nw = max_of(mw, default_w);
		else if (! nw) nw = 200;
		int mh;
		int nh;
		get_preferred_height_for_width(nw, mh, nh);
		if (default_h) nh = max_of(mh, default_h);
		else if (! nh) nh = 200;
		create(title, nw, nh);
	}
	glutShowWindow();
	if (! visible) {
		visible = true;
		visible_count ++;
	}
}

void Window::hide()
{
	if (int wid = get_win_id()) {
		glutSetWindow(wid);
		glutHideWindow();
		if (visible) {
			visible = false;
			visible_count --;
			if (! visible_count) { // all window invisible...
				destroy();
				exit(0);
			}
		}
	}
}

bool Window::is_realized() const
{
	return get_win_id();
}

int Window::get_id() const
{
	return get_win_id();
}
